from ._binding import bind_state_manager  # noqa: F401
from ._built_in_type import (  # noqa: F401
    BuiltInValueType,
)
from ._custom_type import CustomType, JSONType  # noqa: F401
from ._dependency_graph import (  # noqa: F401
    ComputeNode,
    DependencyGraph,
    Input,
    Intermediate,
    Output,
    intermediate,
    output,
)
from ._state_manager import StateManager  # noqa: F401



"""
Werkit's compute wrapper provides a structured, framework-agnostic execution
environment for running any Python code and validating its result using
[1schema][] (or plain JSON Schema). It automatically catches and serializes
errors, too.

By plugging in a werkit.compute.Destination object, you can upon completion send
the result to a queue (or any other destination, such at S3).

[1schema]: https://github.com/metabolize/1schema/
"""

from ._destination import Destination  # noqa: F401
from ._manager import Manager  # noqa: F401
from ._schema import Schema  # noqa: F401



import os
from werkit.aws_lambda.build import (
    collect_zipfile_contents,
    create_venv_with_dependencies,
    create_zipfile_from_dir,
    export_poetry_requirements,
)
from werkit.aws_lambda.deploy import perform_create, perform_update_code


def prepare_zip_file(build_dir, path_to_orchestrator_zip):
    if os.path.isdir(build_dir):
        raise ValueError(f"build_dir should not exist: {build_dir}")

    os.makedirs(build_dir)

    venv_dir = os.path.join(build_dir, "venv")
    zip_dir = os.path.join(build_dir, "zip")
    exported_requirements = os.path.join(build_dir, "requirements.txt")
    export_poetry_requirements(
        output_file=exported_requirements,
        extras=["lambda_common"],
        # This test needs `with_hashes=False` to pass.
        # https://app.circleci.com/pipelines/github/metabolize/werkit/1763/workflows/05f23d1c-8e61-47c5-8c3b-a89fed1a752c/jobs/5181
        with_hashes=False,
    )

    create_venv_with_dependencies(
        venv_dir, install_requirements_from=[exported_requirements]
    )
    collect_zipfile_contents(
        target_dir=zip_dir, venv_dir=venv_dir, src_files=[], src_dirs=["werkit"]
    )
    create_zipfile_from_dir(dir_path=zip_dir, path_to_zipfile=path_to_orchestrator_zip)


def deploy_orchestrator(
    aws_region,
    build_dir,
    path_to_orchestrator_zip,
    orchestrator_function_name,
    role,
    worker_function_name,
    worker_timeout,
    s3_code_bucket=None,
    orchestrator_timeout=600,
    verbose=False,
):
    prepare_zip_file(build_dir, path_to_orchestrator_zip)
    env_vars = {"LAMBDA_WORKER_FUNCTION_NAME": worker_function_name}
    if worker_timeout:
        env_vars["LAMBDA_WORKER_TIMEOUT"] = str(worker_timeout)

    perform_create(
        aws_region=aws_region,
        local_path_to_zipfile=path_to_orchestrator_zip,
        handler="werkit.orchestrator.orchestrator_lambda.handler.handler",
        function_name=orchestrator_function_name,
        role=role,
        timeout=orchestrator_timeout,
        memory_size=3008,  # maximum lambda memory
        env_vars=env_vars,
        s3_code_bucket=s3_code_bucket,
        verbose=verbose,
    )


def update_orchestrator_code(
    aws_region,
    build_dir,
    path_to_orchestrator_zip,
    orchestrator_function_name,
    s3_code_bucket=None,
    verbose=False,
):
    prepare_zip_file(build_dir, path_to_orchestrator_zip)

    perform_update_code(
        aws_region=aws_region,
        local_path_to_zipfile=path_to_orchestrator_zip,
        function_name=orchestrator_function_name,
        s3_code_bucket=s3_code_bucket,
        verbose=verbose,
    )



# Force some libraries to be imported right away. This keeps the execution time
# of the first Lambda invocation consistent with the subsequent execution.
# Otherwise developers users would have to put a big buffer into the timeout
import rds_graphile_worker_client  # noqa: F401
from werkit.compute import Destination


_graphile_worker_client = None


def get_graphile_worker_client(debug=False):
    import os
    from rds_graphile_worker_client import RdsGraphileWorkerClient

    global _graphile_worker_client
    if not _graphile_worker_client:  # pragma: no cover
        config_kwargs = {
            "aws_region": os.environ["AWS_REGION"],
            "pg_username": os.environ["RDS_GRAPHILE_WORKER_PG_USERNAME"],
            "pg_hostname": os.environ["RDS_GRAPHILE_WORKER_PG_HOSTNAME"],
            "pg_port": os.environ["RDS_GRAPHILE_WORKER_PG_PORT"],
            "pg_dbname": os.environ["RDS_GRAPHILE_WORKER_PG_DBNAME"],
        }
        if debug:
            print(f"config_kwargs: {config_kwargs}")
        _graphile_worker_client = RdsGraphileWorkerClient(**config_kwargs)

    return _graphile_worker_client


class RdsGraphileWorkerDestination(Destination):
    def __init__(self, function_name, debug=False):
        self.function_name = function_name
        self.debug = debug

    # TODO: Use a cached property. Tricky because we're using a fork of
    # cachedproperty.
    @property
    def queue_client(self):
        return get_graphile_worker_client(debug=self.debug)

    def debug_queue_message(self):  # pragma: no cover
        cursor = self.queue_client.conn.cursor()
        try:
            cursor.execute("SELECT * from graphile_worker.jobs")
            column_names = [desc[0] for desc in cursor.description]
            values = cursor.fetchall()
        finally:
            cursor.close()
        result = [dict(zip(column_names, these_values)) for these_values in values]
        print(result)

    def send(self, message_key, output_message):
        response = self.queue_client.quick_add_job(self.function_name, output_message)
        print("rds response", response)

        if self.debug:  # pragma: no cover
            self.debug_queue_message()



import datetime
import os
import time


def env_flag(env_var, default):
    environ_string = os.environ.get(env_var, "").strip().lower()
    if not environ_string:
        return default
    return environ_string in ["1", "true", "yes", "on"]


DELAY_SECONDS = int(os.environ.get("DELAY_SECONDS", "0"))
SHOULD_THROW = env_flag("SHOULD_THROW", False)


def serialize_result(message_key, result):
    """
    Simulate the effect of `werkit.Manager()`, which is not used here so this
    test service can be kept to one file.

    TODO: Improve this example by consuming a real schema.
    """
    return {
        "message_key": message_key,
        "success": True,
        "result": result,
        "error": None,
        "error_origin": None,
        "start_time": datetime.datetime(2019, 12, 31).astimezone().isoformat(),
        "duration_seconds": 1.0,
        "runtime_info": None,
    }


def handler(event, context):
    print("event", event)

    time.sleep(DELAY_SECONDS)
    if SHOULD_THROW:
        raise Exception("Whoops!")

    result = event["base"] ** event["exponent"]

    serialized = serialize_result(message_key=event["message_key"], result=result)
    print("serialized", serialized)
    return serialized



import numbers
import typing as t

try:
    from typing import TypeGuard
except ImportError:
    # TODO: Remove when Python 3.9 is dropped.
    from typing_extensions import TypeGuard


BuiltInValueType = t.Union[t.Type[bool], t.Type[int], t.Type[float], t.Type[str]]


def is_built_in_value_type(_type: t.Type) -> TypeGuard[BuiltInValueType]:
    return _type in (bool, int, float, numbers.Number, str)


def coerce_value_to_builtin_type(
    name: str, value_type: BuiltInValueType, value: t.Any
) -> t.Any:
    if not is_built_in_value_type(value_type):
        raise ValueError(
            "Expected built_in_value_type to be a valid built-in value type"
        )
    elif type(value) is value_type:
        return value
    else:
        raise ValueError(
            f"{name} should be type {value_type.__name__}, not {type(value).__name__}"
        )
